//-----------------------------------------------------------------------------
// Mfile - list open handles or file system information
// Command line version of Magnifile C# application.
// Author:  DLang   2009   
//-----------------------------------------------------------------------------

#include <iostream>
#include <iomanip>
#include <assert.h>

#include "ListOpenHandles.h"
extern void DumpFileStats(int nArg, const char* pArg[]);
extern void DisplayFileAllocation(const char* filePathOrPattern);
extern int DisplaySmartInfo(int driveN);
extern  bool DisplayDiskGeometry(int driveN);
extern bool NtfsSysFiles();
extern bool GetNetworkConnections();
extern void ListNetworkConnections(std::ostream& out);

using  namespace std;

const char usage[] =
    "Mfile v3.7 By DLang   " __DATE__ "\n"
    "\n"
    " List Open Handles and File System information\n"
    "\n"
    "File Handles:\n"
    "    -p processName ; Limit file list to specified processName \n"
    "    -y s|m|f|e|p|a ; Report s=summary, m=modules, f=files, e=events, p=procInfo, a=all \n"
    "    -r <repeat#>   ; Repeat output after waitMinutes \n"
    "    -w <waitMin>   ; WaitMinutes defaults to 0.1 \n"
    "\n"
    "Other commands:\n"
    "    -a <filename>  ; Display file cluster allocation \n"
    "    -g <drive#>    ; Display disk geometry \n"
    "    -c             ; Display network connections \n"
    "    -s <drive#>    ; Display smart drive info \n"
    "    -z <drivePath> ; Display File System Stats\n"
    "\n"
    "Example:\n"
    "   mfile                   ; default to list open files for all processes\n"
    "   mfile -y a              ; list all open handles for all processes\n"
    "   mfile -p foobar -y s    ; list handle summary for process foobar \n"
    "   mfile -p foobar -y s -r 9999 -w 0.1  ; list handle summary for process foobar, every .1 minutes \n"
    "\n"
    "   mfile -z                ; list file system information\n"
    "   mfile -z \\\\.\\d       ; list file system information\n"
    "   mfile -a report.xls     ; list file allocation for file report.xls \n"
    "   mfile -a *.xls          ; \n"
    "\n";

/* ------------------------------------------------------------------------- */
#ifndef HAVE_GETOPT
const char *optarg = NULL; /* argument associated with option */
int opterr   = 0;       /* if error message should be printed */
int optind   = 1;       /* index into parent argv vector */
int optopt   = 0;       /* character checked for validity */

#define BADCH (int)'?'
#define EMSG ""

/* ------------------------------------------------------------------------- */
int getopt(int nargc, const char * nargv[], const char *ostr)
{
	static const char* place = EMSG; /* option letter processing */
	const char* oli; /* option letter list index */
	const char* p;
	
	if (!*place) 
    { /* update scanning pointer */
		if (optind >= nargc || *(place = nargv[optind]) != '-') 
        {
			place = EMSG;
			return (EOF);
		}
		if (place[1] && *++place == '-') 
        { /* found "--" */
			++optind;
			place = EMSG;
			return (EOF);
		}
	} /* option letter okay? */

	if ((optopt = (int)*place++) == (int)':' ||
		!(oli = strchr(ostr, optopt))) 
    {
		/*
		* For backwards compatibility: don't treat '-' as an
		* option letter unless caller explicitly asked for it.
		*/
		if (optopt == (int)'-')
			return (EOF);
		if (!*place)
			++optind;
		if (opterr) {
			if (!(p = strrchr(*nargv, '/')))
				p = *nargv;
			else
				++p;
			fprintf(stderr, "%s: illegal option -- %c\n", p, optopt);
		}
		return (BADCH);
	}

	if (*++oli != ':') 
    { /* don't need argument */
		optarg = NULL;
		if (!*place)
			++optind;
	} else { /* need an argument */
		if (*place) /* no white space */
			optarg = place;
		else if (nargc <= ++optind) 
        { /* no arg */
			place = EMSG;
			if (!(p = strrchr(*nargv, '/')))
				p = *nargv;
			else
				++p;
			if (opterr)
				fprintf(stderr, "%s: option requires an argument -- %c\n",
				p, optopt);
			return (BADCH);
		} else /* white space */
			optarg = nargv[optind];
		place = EMSG;
		++optind;
	}

	return (optopt); /* dump back option letter */
}
#endif
   
//-----------------------------------------------------------------------------
inline void ToggleBool(bool& b, const char* arg, char value)
{
    b  = (strchr(arg, value) != NULL) ? !b : b;
}

//-----------------------------------------------------------------------------
int main(int argc, const char* argv[])
{
    ListOpenHandles listOpenHandles;
    
    int c;
    while ((c = getopt(argc, argv, "a:cg:np:r:s:w:y:z?h")) != -1)
    {
        switch (c)
        {
          case 'a':     // report file allocation
            DisplayFileAllocation(optarg);
            return 0;

          case 'c':     // display network connections
            ListNetworkConnections(std::cout);
            return 0;

          case 'g':
            DisplayDiskGeometry(atoi(optarg));
            return 0;

          case 'r':     // repeat count (File Performance command)
            listOpenHandles.m_repeat = atoi(optarg);
            break;

          case 'w':     // waitMinutes (File Performance command)
            listOpenHandles.m_waitMin = atof(optarg);
            break;

          case 'n':
            NtfsSysFiles();
            return 0;

          case 'p':
            listOpenHandles.m_process = optarg;
            break;

          case 's': // -s <drive#>, ex:  -s 0
            DisplaySmartInfo(atoi(optarg));
            return 0;

          case 'y':
            ToggleBool(listOpenHandles.m_showSummary, optarg, 's');
            ToggleBool(listOpenHandles.m_showModules, optarg, 'm');
            ToggleBool(listOpenHandles.m_showFiles, optarg, 'f');
            ToggleBool(listOpenHandles.m_showEvents, optarg, 'e');
            ToggleBool(listOpenHandles.m_showAll, optarg, 'a');
            ToggleBool(listOpenHandles.m_showHex, optarg, 'x');
            ToggleBool(listOpenHandles.m_showProcInfo, optarg, 'p');
            break;

          case 'z':
            return DumpFileStats(argc-optind, argv+optind);

          default:
            cout << usage << endl;
            return -1;
        }
    }

    int resultCode = listOpenHandles.ReportOpenFiles(cout);
	return resultCode;
}

